# Permanent Digital Twin of a harvester

With the fml40 python reference implementation package, we have developed a Digital Twin of a harvester, which is modeled based on the ForestML 4.0 language and runs permanently. The Digital Twin is implemented using the fog approach, i.e., it has an edge DT part (living on our internal servers) and a cloud DT part (within the S³I Repository). The DT is reachable via S³I-B messages. To communicate with the DT, we provide a python script named `hmi.py` in the folder `demo`. Please feel free to contact us in order to get the necessary credentials file for this HMI. Put the credentials in the folder `config` and  you can run `hmi.py`. This documentation introduces various service functions the Digital Twin provides and show its JSON-based structure within the S³I Directory and Repository.  


## S³I Directory entry
S³I Directory provides a thing's metadata, i.e., what it looks like and how it is reachable (endpoints) - just like the yellow pages. S³I Directory provides its own meta model the fml40 data model is mapped to. In order to query the corresponding JSON data, you can send an HTTP request to the S³I Directory (see [this S³I document](https://kwh40.pages.rwth-aachen.de/-/s3i/-/jobs/1095388/artifacts/public/html/md/communication.html)). The Digital Twin of the harvester has the following directory entry (note the two endpoints: one for S³I-B and one for S³I Repository):
```
{
    "thingId": "s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03",
    "policyId": "s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03",
    "attributes": {
        "ownedBy": "318b5bc4-7479-4d8d-b347-1019ca194d81",
        "class": "ml40::Thing",
        "allEndpoints": [
            "s3ib://s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03",
            "https://ditto.s3i.vswf.dev/api/2/things/s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03/"
        ],
        "type": "component",
        "name": "mmi_fog_harvester",
        "location": {
            "longitude": 6.45435,
            "latitude": 52.543534
        },
        "dataModel": "fml40",
        "thingStructure": {
            "class": "ml40::Thing",
            "links": [
                {
                    "association": "roles",
                    "target": {
                        "class": "fml40::Harvester"
                    }
                },
                {
                    "association": "features",
                    "target": {
                        "class": "fml40::ProvidesProductionData"
                    }
                },
                {
                    "association": "features",
                    "target": {
                        "class": "ml40::ProductionData"
                    }
                },
                {
                    "association": "features",
                    "target": {
                        "class": "fml40::AcceptsFellingJobs"
                    }
                },
                {
                    "association": "features",
                    "target": {
                        "class": "fml40::Harvests"
                    }
                },
                {
                    "association": "features",
                    "target": {
                        "class": "ml40::JobList"
                    }
                },
                {
                    "association": "features",
                    "target": {
                        "class": "ml40::Composite",
                        "links": [
                            {
                                "class": "ml40::Thing",
                                "links": [
                                    {
                                        "association": "roles",
                                        "target": {
                                            "class": "ml40::Engine"
                                        }
                                    },
                                    {
                                        "association": "features",
                                        "target": {
                                            "class": "ml40::RotationalSpeed"
                                        }
                                    }
                                ]
                            },
                            {
                                "class": "ml40::Thing",
                                "links": [
                                    {
                                        "association": "roles",
                                        "target": {
                                            "class": "ml40::Crane"
                                        }
                                    },
                                    {
                                        "association": "features",
                                        "target": {
                                            "class": "ml40::Shared"
                                        }
                                    }
                                ]
                            }
                        ]
                    }
                }
            ]
        }
    }
}
```

## S³I Repository entry
In contrast to the S³I Directory's metadata, fml40 data is directly mapped to JSON when using a run time environment like an edge device or S³I Repository that contains the DT's data itself. The repository entry can be queried via HTTP REST request, refer to [S³I Documentation](https://kwh40.pages.rwth-aachen.de/-/s3i/-/jobs/1095388/artifacts/public/html/md/communication.html). The permanent Digital Twin of the harvester has a repository entry which is shown in the following:
```
{
    "thingId": "s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03",
    "policyId": "s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03",
    "attributes": {
        "class": "ml40::Thing",
        "name": "mmi_fog_harvester",
        "roles": [
            {
                "class": "fml40::Harvester"
            }
        ],
        "features": [
            {
                "class": "fml40::ProvidesProductionData"
            },
            {
                "class": "ml40::ProductionData",
                "subFeatures": [
                    {
                        "class": "ml40::Composite",
                        "targets": [
                            {
                                "class": "ml40::Thing",
                                "name": "Stammsegment 4711",
                                "roles": [
                                    {
                                        "class": "fml40::StemSegment"
                                    }
                                ],
                                "features": [
                                    {
                                        "class": "ml40::Location",
                                        "latitude": 50.1231,
                                        "longitude": 6.1231
                                    },
                                    {
                                        "class": "fml40::StemSegmentProperties",
                                        "diameter": 0.4,
                                        "length": 2,
                                        "woodType": "Spruce"
                                    }
                                ]
                            },
                            {
                                "class": "ml40::Thing",
                                "name": "Stammsegment 4712",
                                "roles": [
                                    {
                                        "class": "fml40::StemSegment"
                                    }
                                ],
                                "features": [
                                    {
                                        "class": "ml40::Location",
                                        "latitude": 50.1231,
                                        "longitude": 6.1231
                                    },
                                    {
                                        "class": "fml40::StemSegmentProperties",
                                        "diameter": 0.6,
                                        "length": 3,
                                        "woodType": "Spruce"
                                    }
                                ]
                            }
                        ]
                    }
                ]
            },
            {
                "class": "fml40::AcceptsFellingJobs"
            },
            {
                "class": "fml40::Harvests"
            },
            {
                "class": "ml40::JobList"
            },
            {
                "class": "ml40::Composite",
                "targets": [
                    {
                        "class": "ml40::Thing",
                        "name": "Motor",
                        "roles": [
                            {
                                "class": "ml40::Engine"
                            }
                        ],
                        "features": [
                            {
                                "class": "ml40::RotationalSpeed",
                                "rpm": 2191
                            }
                        ]
                    },
                    {
                        "class": "ml40::Thing",
                        "name": "Kran",
                        "roles": [
                            {
                                "class": "ml40::Crane"
                            }
                        ],
                        "features": [
                            {
                                "class": "ml40::Shared",
                                "targets": [
                                    {
                                        "class": "ml40::Thing",
                                        "name": "Harvesterkopf",
                                        "roles": [
                                            {
                                                "class": "fml40::HarvestingHead"
                                            }
                                        ],
                                        "features": [
                                            {
                                                "class": "fml40::Cuts"
                                            },
                                            {
                                                "class": "fml40::Grabs"
                                            }
                                        ]
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    },
    "features": {}
}
```

## Service Call
As shown in the directory entry, the DT provides several service functions, which is summarized here:
```
class: fml40::AcceptsFellingJobs
service functions:
    - acceptJob(job)
    - queryJobStatus(identifier)
    - removeJob(identifier)

class: fml40::ProvidesProductionData
service functions:
    - getProductionData(name)
```
In order to call remotely them, we need to send an S³I-B ServiceRequest message to the Digital Twin via S³I Broker. The class/service function name and parameters need to be declared within the message.
The following example shows how to call `removeJob` from class `fml40::AcceptsFellingJob`
```
{
    "identifier": "s3i:94f8c4d1-9d6c-45e8-a137-3bb33b1004b6" # message ID
    "sender": "s3i:7f46a255-3245-439e-84f0-090f0b221965", # ID of my private HMI
    "receivers": ["s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03"], # ID of the receiving Digital Twin
    "serviceType": "fml40::AcceptsFellingJob/removeJob", # specifies the service function and class, which should be called
    "parameters": {"identifier": "s3i:3ba16ab9-a472-42b7-af97-c19acef55196"}, # specifies the input to this service function
    "messageType": "ServiceRequest", # message type
    "replyToEndpoint": "s3ib://s3i:7f46a255-3245-439e-84f0-090f0b221965" # Endpoint of my private HMI,
}
```
The message can be sent directly via [the S³I Broker API](https://broker.s3i.vswf.dev/apidoc) or using [S³I Python Lib](https://kwh40.pages.rwth-aachen.de/-/s3i/-/jobs/1095388/artifacts/public/html/S3I-Package.html). An example can be found [here](https://git.rwth-aachen.de/kwh40/s3i/-/blob/master/demo/demo2_sender_hmi.py)

## Query Value
The DT can also handle S³I GetValueRequest messages like this example for querying its engine's rpm:
```
{
    'sender': 's3i:7f46a255-3245-439e-84f0-090f0b221965',   # ID of my private HMI
    'identifier': 's3i:23c99920-c369-4074-83b9-17beb84bb2d2', # message ID
    'receivers': ['s3i:b6d1cc6d-896c-40fe-9403-b5b7682b1d03'], # ID of the digital Twin
    'messageType': 'getValueRequest', # message Type
    'replyToEndpoint': 's3ibs://s3i:7f46a255-3245-439e-84f0-090f0b221965', # Endpoint of my private HMI
    'attributePath': 'attributes/features/ml40::Composite/targets/ml40::Engine/features/ml40::RotationalSpeed/rpm' # the path of rpm value
}
```
You can also query the whole JSON file and, in this case, you just need to replace the `attributePath` with an empty string.