#!/usr/bin/env python
"""launch_dt is a script for launching digital twins.  It proivdes
additional function for receiving informations about available twins
and functionalities.

"""

import argparse
import os
from pathlib import Path

from ml.dt_factory import create_thing
from ml.tools import load_config

root_path = os.path.abspath("")


def create_argparser():
    parent_parser = argparse.ArgumentParser(prog="", description="FML40 argument parse")

    sub_parsers = parent_parser.add_subparsers(dest="command")

    launch_parser = sub_parsers.add_parser("launch")
    launch_parser.add_argument("dt_config", nargs=1, type=str, help="path of the configuration file for the digital twin")

    list_parser = sub_parsers.add_parser("list")
    list_parser.add_argument("ml", nargs=1,
                             choices=["dts"], # Todo list all ml and fml40 functionalities and properties
                             help="list the current digital twins configured")

    return parent_parser


def main():
    """Executes actions depending on user input.
    """
    parser = create_argparser()
    args = parser.parse_args()

    if args.command == "list":
        if args.ml[0] == "dts":
            dt_names = []
            config_paths = Path("configs").glob("*.json")
            for config_path in config_paths:
                try:
                    # TODO: Validate json schemas
                    config = load_config(config_path)
                    dt_names.append(config["attributes"]["name"])
                except:
                    pass
            print(dt_names)
    elif args.command == "launch":
        dt_config = args.dt_config[0]
        try:
            model = load_config(os.path.join(root_path, "configs", dt_config))
        except FileNotFoundError:
            print("{} can not be found".format(dt_config))
            return
        secret = input("Please enter the client secret of the digital twin with id {}: ".format(model["thingId"]))
        is_broker = input("Does this DT has a S³I-B endpoint? (j/n) ")
        if "j" == is_broker.lower():
            is_broker = True
        elif "n" == is_broker.lower():
            is_broker = False
        is_repo = input("Does this DT has a cloud copy in the S³I Repository? (j/n) ")
        if "j" == is_repo.lower():
            is_repo = True
        elif "n" == is_repo.lower():
            is_repo = False

        dt = create_thing(model_json=model, grant_type="client_credentials",
                           oauth2_secret=secret,
                           is_broker_rest=True,
                           is_broker=True,
                           is_repository=True)
        dt.run_forever()


if __name__ == "__main__":
    main()
